use std::marker::PhantomData;

use super::*;

/// [`Mode`] equivalent of [`ParseResult`].
pub type ModeResult<M, A, E, I> = Result<<M as Mode>::ParseSuccess<A, I>, E>;

/// Mode of parsing.
///
/// | [`Mode`]         | [`Mode::ParseSuccess<A, I>`] | [`Mode::ExtensionResult<A, E>`] | [`Mode::ExtensionSource<A>`] |
/// |------------------|------------------------------|---------------------------------|------------------------------|
/// | [`RegularMode`]  | `A`                          | [`Result<A, E>`]                | `A`                          |
/// | [`InliningMode`] | `(A, I)`                     | `E`                             | [`()`]                       |
///
/// [`InliningMode`]: crate::rstd::inlining::InliningMode
/// [`()`]: unit
pub trait Mode {
    /// Successful parsing, may countain the parser itself
    /// (`I`, usually [`Inlining`]).
    ///
    /// See [`FactoryParse::deserialize`].
    type ParseSuccess<A, I>;

    /// Result of extending the value, failing sometimes or always.
    ///
    /// See [`FactoryParse::extend`].
    type ExtensionResult<A, E>;

    /// Data enough to try extending the value.
    ///
    /// May be empty for always-failing extensions.
    ///
    /// See [`FactoryParse::extend`].
    type ExtensionSource<A>;

    fn bind<A0, A1, E, I>(
        s: Self::ParseSuccess<A0, I>,
        f: impl FnOnce(A0) -> Result<A1, E>,
    ) -> ModeResult<Self, A1, E, I>;

    fn map<A0, A1, I>(
        s: Self::ParseSuccess<A0, I>,
        f: impl FnOnce(A0) -> A1,
    ) -> Self::ParseSuccess<A1, I>;

    /// Discard any extra information contained in [`Mode::ParseSuccess`].
    fn seal<A, I>(s: Self::ParseSuccess<A, I>) -> A;

    fn xmap_err<A, E0, E1>(
        result: Self::ExtensionResult<A, E0>,
        f: impl FnOnce(E0) -> E1,
    ) -> Self::ExtensionResult<A, E1>;

    fn xbind<A0, A1, E>(
        result: Self::ExtensionResult<A0, E>,
        f: impl FnOnce(A0) -> Result<A1, E>,
    ) -> Self::ExtensionResult<A1, E>;

    /// Convert [`Mode::ExtensionResult`] to [`Result`].
    fn xseal<A, E>(result: Self::ExtensionResult<A, E>) -> Result<A, E>;

    fn smap<A0, A1>(
        source: Self::ExtensionSource<A0>,
        f: impl FnOnce(A0) -> A1,
    ) -> Self::ExtensionSource<A1>;

    fn prepare<A>(a: A) -> Self::ExtensionSource<A>;

    /// For abstract [`FactoryParse::extend`] implementations.
    fn xsbind<AB, A, B, E>(
        ab: Self::ExtensionSource<AB>,
        t2ab: impl FnOnce(AB) -> (A, B),
        ce: impl FnOnce(Self::ExtensionSource<B>) -> Self::ExtensionResult<B, E>,
        ab2t: impl FnOnce(A, B) -> Result<AB, E>,
    ) -> Self::ExtensionResult<AB, E>;
}

pub trait ParseMode {
    type Mode: ?Sized + Mode;
}

pub trait ParseModeExt: ParseMode {
    fn bind<A0, A1, E, I>(
        s: <Self::Mode as Mode>::ParseSuccess<A0, I>,
        f: impl FnOnce(A0) -> Result<A1, E>,
    ) -> ModeResult<Self::Mode, A1, E, I> {
        <Self::Mode as Mode>::bind(s, f)
    }

    fn map<A0, A1, I>(
        s: <Self::Mode as Mode>::ParseSuccess<A0, I>,
        f: impl FnOnce(A0) -> A1,
    ) -> <Self::Mode as Mode>::ParseSuccess<A1, I> {
        <Self::Mode as Mode>::map(s, f)
    }

    fn smap<A0, A1>(
        source: <Self::Mode as Mode>::ExtensionSource<A0>,
        f: impl FnOnce(A0) -> A1,
    ) -> <Self::Mode as Mode>::ExtensionSource<A1> {
        <Self::Mode as Mode>::smap(source, f)
    }
}

impl<F: ParseMode> ParseModeExt for F {}

pub trait WithParseMode: ParseMode {
    type WithMode: ?Sized;
}

pub trait ImplMode {
    type Mode: ?Sized + Mode;
}

impl<T: ?Sized + ImplMode> ParseMode for T {
    type Mode = <Self as ImplMode>::Mode;
}

impl<T: ?Sized + ImplMode> WithParseMode for T {
    type WithMode = WithMode<Self, <Self as ParseMode>::Mode>;
}

pub struct WithMode<T: ?Sized, M: ?Sized>(PhantomData<M>, T);

impl<'a, Ctx: Context<'a>, F: FactoryModeParse<'a, Ctx>> FactoryParse<'a, Ctx> for F {
    fn deserialize(&self, inctx: impl InCtx<'a, Ctx>) -> ParseResult<'a, Ctx, Self> {
        self.mdeserialize(inctx)
            .map(<<Self as ParseMode>::Mode as Mode>::seal)
    }

    fn extend(&self, mentionable: Self::Mtbl, tail: &[u8]) -> ParseResult<'a, Ctx, Self> {
        <<Self as ParseMode>::Mode as Mode>::xseal(self.mextend(
            <<Self as ParseMode>::Mode as Mode>::prepare(mentionable),
            tail,
        ))
    }
}

impl<'a, Ctx: Context<'a>, F: FactoryParse<'a, Ctx> + ParseMode> Factory<'a, Ctx> for F
where
    F::Mtbl: MentionableTop<'a, Ctx>,
{
    type _Mtbl = Self::Mtbl;
}

pub type ParseResultM<'a, Ctx, F, I> =
    ModeResult<<F as ParseMode>::Mode, Mtbl<'a, Ctx, F>, ParseError<'a, Ctx, F>, I>;

pub type ExtensionResultM<'a, Ctx, F> =
    <<F as ParseMode>::Mode as Mode>::ExtensionResult<Mtbl<'a, Ctx, F>, ParseError<'a, Ctx, F>>;

pub type ExtensionSourceM<'a, Ctx, F> =
    <<F as ParseMode>::Mode as Mode>::ExtensionSource<Mtbl<'a, Ctx, F>>;

pub trait FactoryModeParse<'a, Ctx: Context<'a>>: FactoryBase<'a, Ctx> + ParseMode {
    fn mdeserialize<I: InCtx<'a, Ctx>>(&self, inctx: I) -> ParseResultM<'a, Ctx, Self, I>;

    fn mextend(
        &self,
        mentionable: ExtensionSourceM<'a, Ctx, Self>,
        tail: &[u8],
    ) -> ExtensionResultM<'a, Ctx, Self>;
}

pub trait FactoryModeProxy<'a, Ctx: Context<'a>> {
    type F: FactoryBase<'a, Ctx> + ParseMode;

    fn pmdeserialize<I: InCtx<'a, Ctx>>(f: &Self::F, inctx: I)
        -> ParseResultM<'a, Ctx, Self::F, I>;

    fn pmextend(
        f: &Self::F,
        mentionable: ExtensionSourceM<'a, Ctx, Self::F>,
        tail: &[u8],
    ) -> ExtensionResultM<'a, Ctx, Self::F>;
}

impl<'a, Ctx: Context<'a>, F: FactoryBase<'a, Ctx> + WithParseMode> FactoryModeParse<'a, Ctx> for F
where
    F::WithMode: FactoryModeProxy<'a, Ctx, F = F>,
{
    fn mdeserialize<I: InCtx<'a, Ctx>>(&self, inctx: I) -> ParseResultM<'a, Ctx, Self, I> {
        <F::WithMode as FactoryModeProxy<'a, Ctx>>::pmdeserialize(self, inctx)
    }

    fn mextend(
        &self,
        mentionable: ExtensionSourceM<'a, Ctx, Self>,
        tail: &[u8],
    ) -> ExtensionResultM<'a, Ctx, Self> {
        <F::WithMode as FactoryModeProxy<'a, Ctx>>::pmextend(self, mentionable, tail)
    }
}

pub struct RegularMode;

impl Mode for RegularMode {
    type ParseSuccess<A, I> = A;

    type ExtensionResult<A, E> = Result<A, E>;

    type ExtensionSource<A> = A;

    fn bind<A0, A1, E, I>(
        s: Self::ParseSuccess<A0, I>,
        f: impl FnOnce(A0) -> Result<A1, E>,
    ) -> ModeResult<Self, A1, E, I> {
        f(s)
    }

    fn map<A0, A1, I>(
        s: Self::ParseSuccess<A0, I>,
        f: impl FnOnce(A0) -> A1,
    ) -> Self::ParseSuccess<A1, I> {
        f(s)
    }

    fn seal<A, I>(s: Self::ParseSuccess<A, I>) -> A {
        s
    }

    fn xmap_err<A, E0, E1>(
        result: Self::ExtensionResult<A, E0>,
        f: impl FnOnce(E0) -> E1,
    ) -> Self::ExtensionResult<A, E1> {
        result.map_err(f)
    }

    fn xbind<A0, A1, E>(
        result: Self::ExtensionResult<A0, E>,
        f: impl FnOnce(A0) -> Result<A1, E>,
    ) -> Self::ExtensionResult<A1, E> {
        result.and_then(f)
    }

    fn xseal<A, E>(result: Self::ExtensionResult<A, E>) -> Result<A, E> {
        result
    }

    fn smap<A0, A1>(
        source: Self::ExtensionSource<A0>,
        f: impl FnOnce(A0) -> A1,
    ) -> Self::ExtensionSource<A1> {
        f(source)
    }

    fn prepare<A>(a: A) -> Self::ExtensionSource<A> {
        a
    }

    fn xsbind<AB, A, B, E>(
        ab: Self::ExtensionSource<AB>,
        t2ab: impl FnOnce(AB) -> (A, B),
        ce: impl FnOnce(Self::ExtensionSource<B>) -> Self::ExtensionResult<B, E>,
        ab2t: impl FnOnce(A, B) -> Result<AB, E>,
    ) -> Self::ExtensionResult<AB, E> {
        let (b, c) = t2ab(ab);
        ab2t(b, ce(c)?)
    }
}

pub trait QRegularFactory<'a, Ctx: Context<'a>>:
    FactoryBase<'a, Ctx> + ImplMode<Mode = RegularMode>
{
    fn qrdeserialize(&self, inctx: impl InCtx<'a, Ctx>) -> ParseResult<'a, Ctx, Self>;
    fn qrextend(&self, mentionable: Self::Mtbl, tail: &[u8]) -> ParseResult<'a, Ctx, Self>;
}

pub trait RegularFactory<'a, Ctx: Context<'a>>:
    FactoryBase<'a, Ctx> + ParseMode<Mode = RegularMode>
{
    fn rdeserialize(&self, inctx: impl InCtx<'a, Ctx>) -> ParseResult<'a, Ctx, Self>;
    fn rextend(&self, mentionable: Self::Mtbl, tail: &[u8]) -> ParseResult<'a, Ctx, Self>;
}

impl<'a, Ctx: Context<'a>, F: FactoryModeParse<'a, Ctx> + ParseMode<Mode = RegularMode>>
    RegularFactory<'a, Ctx> for F
{
    fn rdeserialize(&self, inctx: impl InCtx<'a, Ctx>) -> ParseResult<'a, Ctx, Self> {
        self.mdeserialize(inctx)
    }

    fn rextend(&self, mentionable: Self::Mtbl, tail: &[u8]) -> ParseResult<'a, Ctx, Self> {
        self.mextend(mentionable, tail)
    }
}

impl<'a, Ctx: Context<'a>, F: QRegularFactory<'a, Ctx>> FactoryModeProxy<'a, Ctx>
    for WithMode<F, RegularMode>
{
    type F = F;

    fn pmdeserialize<I: InCtx<'a, Ctx>>(f: &Self::F, inctx: I) -> ParseResultM<'a, Ctx, F, I> {
        f.qrdeserialize(inctx)
    }

    fn pmextend(f: &F, mentionable: Mtbl<'a, Ctx, F>, tail: &[u8]) -> ExtensionResultM<'a, Ctx, F> {
        f.qrextend(mentionable, tail)
    }
}
